@@ -15,6 +15,10 @@
 # pragma once
 #endif // defined(_MSC_VER) && (_MSC_VER >= 1200)
 
+#if defined(ASIO_ENABLE_TWO_LOCK_QUEUE)
+#include "asio/detail/task_io_service_2lock.hpp"
+#else // defined(ASIO_ENABLE_TWO_LOCK_QUEUE)
+
 #include "asio/detail/push_options.hpp"
 
 #include "asio/error_code.hpp"
@@ -416,4 +420,6 @@ private:
 
 #include "asio/detail/pop_options.hpp"
 
+#endif // defined(ASIO_ENABLE_TWO_LOCK_QUEUE)
+
 #endif // ASIO_DETAIL_TASK_IO_SERVICE_HPP
